#include "Controller.h"
#include <string>
#include <sstream>
#include <iostream>
#include <list>

#include <RL_PP/State.h>
#include <RL_PP/AbstractAction.h>
#include <RL_PP/ActionTileCoding.h>
#include <RL_PP/ActionTilings.h>
#include <RL_PP/ActionLookupTable.h>
#include <RL_PP/Utility.h>


static const bool loadValueFunction = false;



/*
Die Klasse Action ist die einfachste Implementation einer Aktion für die Aufgabe.
Möglich wäre z.B. die torque für das Kommando erst aus einem besseren/abstrakten
 Parameterzu berechnen.
*/
class Action : public AbstractAction
{
  public:
    Action( int t ) : AbstractAction()
    {
      // was sinnvolles machen
    }

    inline const std::string command() const
    {
      std::ostringstream tmp;
      // hier Aktionskommandos generieren
      return tmp.str();
    }

  private:
};



int Controller::theSocket;


Controller::Controller( const int aSocket)
  : Watkins( "rl.conf" )
{
  theSocket = aSocket;
  episodes = 0;

  if ( loadValueFunction ) {
    load();
    // hier die Aktionen setzen
    for( int j=0; j< 1; ++j ) {
      avfPtr->setAction( j, new Action( j ) );
    }
  }
  else {
    State* statePtr = new State();
    // Zustandsvariablen mit MIN- und MAX-Wert registrieren
    stateId = statePtr->registerVariable( "TEST", -10.0, 10.0 );
  
    // ValueFunction anlegen
    // hier "ActionTileCoding"
    avfPtr = new ActionTileCoding( statePtr );
    // hier die Aktionen bei der valueFunction registrieren
    for( int j=0; j<1; ++j ) {
      avfPtr->registerAction( new Action( j ) );
    }
    // Zusandsraum festlegen
    // hier mit TileCodings
    Types::TILINGS_VAR_STRUCT vs;
    std::list< Types::TILINGS_VAR_STRUCT > vsList;

    vs.id = stateId;
    vs.resolution = 2.0; // gewünschte effektive Auflösung
    vs.tileType = Types::CLIP;
    vsList.push_back( vs );
    ActionTilings* t = new ActionTilings( 2, vsList ); //Anzahl der Tiles (2)
    static_cast< ActionTileCoding* >(avfPtr)->registerTiling( t );
  }
}


void Controller::firstInit()
{
  std::cerr << "++++++++ FIRST_INIT +++++++++++\n";

  recv ( theSocket, buffer, 1024, 0 );

  //alle joints in Ausgangslage fahren und fixieren
  sendMsg("(reset)\n");
  sendMsg("(pidmaxtorque** 800)\n");
  sendMsg("(pid* 512 202 800 202 800 512 512 512 512 512 512 512 512 512 512 512 512 512 512)\n");
  for ( int i = 0; i <= 200; ++i ) {
    sendMsg("(done)\n");
    recv ( theSocket, buffer, 1024, 0 );
  }

  sendMsg("(pid 10 562)\n");
  sendMsg("(pid 11 562)\n");
  sendMsg("(pid 18 562)\n");
  for ( int i = 0; i <= 100; ++i ) {
    sendMsg("(done)\n");
    recv ( theSocket, buffer, 1024, 0 );
  }


  sendMsg("(pid 13 452)\n");
  sendMsg("(pid 15 412)\n");
  sendMsg("(pid 17 562)\n");
  for ( int i = 0; i <= 50; ++i ) {
    sendMsg("(done)\n");
    recv ( theSocket, buffer, 1024, 0 );
  }

  sendMsg("(pid 13 412)\n");
  sendMsg("(pid 15 312)\n");
  sendMsg("(pid 17 612)\n");
  for ( int i = 0; i <= 50; ++i ) {
    sendMsg("(done)\n");
    recv ( theSocket, buffer, 1024, 0 );
  }

  sendMsg("(pid 13 352)\n");
  sendMsg("(pid 15 212)\n");
  sendMsg("(pid 17 662)\n");
  sendMsg("(pid 19 562)\n");
  for ( int i = 0; i <= 50; ++i ) {
    sendMsg("(done)\n");
    recv ( theSocket, buffer, 1024, 0 );
  }

  // Snapshot abspeichern
  sendMsg("(save)\n");
  sendMsg("(done)\n");
}




void Controller::init()
{
  recv ( theSocket, buffer, 1024, 0 );
  sendMsg("(reset)\n");
  sendMsg("(done)\n");
  recv ( theSocket, buffer, 1024, 0 );
  sendMsg("(restore)\n");
  sendMsg("(done)\n");
  recv ( theSocket, buffer, 1024, 0 );
  sendMsg("(pidmaxtorque** 800)\n");
  sendMsg("(pid* 512 202 800 202 800 512 512 512 512 562 562 512 352 512 212 512 662 562 562)\n");
  sendMsg("(done)\n");
}


void Controller::run()
{
  // Aufruf der RL++-Kontrollschleife
  firstInit();
  controlLoop();
}



void Controller::setInitialState()
{
  init();
  steps=0;
  ++episodes;

  //Umweltinformationen empfangen
  sprintf(buffer, "");
  int len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';
  sscanf( strchr(buffer, (int)'p'), "pos hip  %f %f %f %*s", &hipX, &hipY, &hipZ );
  sscanf( strchr(buffer, (int)'b'), "ball  %f %f %f %*s", &ballX, &ballY, &ballZ );

  std::cerr << "initial  hipPos = (" << hipX << " , " << hipY << " , " << hipZ << "\n";
  std::cerr << "initial ballPos = (" << ballX << " , " << ballY << " , " << ballZ << "\n";
  // neue Werte der Zustandsvariablen setzen
  avfPtr->getStatePtr()->setValueDbl( stateId, 3.1415 );
}


void Controller::takeAction()
{
  // +++ TEST ++++
  // Schiess Doch!
  sendMsg("(pid 15 600)\n");
  sendMsg("(done)\n");
  return;

  const Action* action = static_cast< const Action* >( avfPtr->getActiveAction() );
  if ( action == NULL ) {
    std::cerr << "USERFEHLER: avfPtr->getActiveAction() liefert NULL !!!\n";
    exit(-1);
  }
  sendMsg(action->command().c_str());
  sendMsg("(done)\n");
}


const bool Controller::observeState()
{
  ++steps;
  sprintf(buffer, "");
  int len = recv ( theSocket, buffer, 1024, 0 );
  buffer[len] = '\0';
  
  // hier sinnvolle Zustandsgroessen auslesen
  
  avfPtr->getStatePtr()->setValueDbl( stateId, 3.1415 );

  if ( steps > 200 ) {
    sendMsg("(done)\n");
    return false;
  }
  return true;
}


const double Controller::calculateReward()
{
  double reward = -0.1;
  return reward;
}


void Controller::finishEpisode()
{
  std::cout << "episodes= " << episodes << " steps = " << steps << "\n";
  // valueFunction speichern
  if ( (episodes%10) == 0 ) {
    save();
  }
}


void Controller::save() const
{
  // save valuefunction
  std::cerr << "saving ... ";
  int res = avfPtr->save( "aufgabe6.vf" );
  if (res==0) {
    std::cerr << "okay\n";
  }
  else {
    std::cerr << "failed\n";
  }
}

void Controller::load()
{
  // load valuefunction
  avfPtr = new ActionTileCoding( "aufgabe6.vf" );
  avfPtr->setEpisodeStateHistory( true );
  stateId= avfPtr->getStatePtr()->getVariableId( "TEST" );
}
