#ifndef _CONTROLLER_H_
#define _CONTROLLER_H_

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <errno.h>

// RL++Lernklasse einbinden
#include <RL_PP/Watkins.h>


// der Controller muss direkt von der dem gewünschten Lernverfahren assoziierten
// Klasse abgeleitet werden
class Controller : public Watkins
{
  public:

    Controller( const int aSocket);
    void init();
    void firstInit();
    void run();

  protected :

    inline static void sendMsg(const char* text);

    // in itialisiert eine neue Lernepisode
    virtual void setInitialState();

    // führt die entsprechend der Auswahlpolitik (z.B. GREEDY) gewählte Aktion aus
    virtual void takeAction();

    // berechnet u.a. aus der Umweltinformation die Zustandsgrößen und übergibt
    // diese dem Lernsystem
    // liefert false, falls die Lernepisode beendet werden soll
    virtual const bool observeState();

    // liefert das Reward-Signal
    // Mit der Wahl des reward steht und fällt der Lernerfolg
    // Faustregel: denken Sie nicht zu kompliziert!
    virtual const double calculateReward();

    // viel Platz für Statistik und Auswertungen
    virtual void finishEpisode();

    // sichert die ValueFunction in eine Datei
    virtual void save() const;

    // lädt die ValueFunction aus einer Datei
    virtual void load();

  private:

    static int theSocket;
    char buffer[1024];

    // die IDs der Zustandsvariablen
    int stateId;

    // Hilfsvariablen
    float hipX, hipY, hipZ;
    float ballX, ballY, ballZ;

    unsigned int steps;
    unsigned int episodes;
  
    int len;

};


inline void Controller::sendMsg(const char* text)
{
  //std::cerr << "sending ... " << std::string(text) << "\n";
  if ( send(theSocket, text, strlen(text), 0) != static_cast<int>(strlen(text))) {
    fprintf(stderr, "send() hat eine unterschiedliche Anzahl von Bytes versendet, als erwartet !!!!: %s\n", strerror(errno));
    exit(-1);
  }
}

#endif
